@@ -228,6 +228,11 @@ struct MVMBoolificationSpec {
     MVMuint32  mode;
 };
 
+/* Constant for incrementing the type cache ID for new STables. This leaves
+ * the lowest bits free for caches to attach flags (of note, the multi
+ * dispatch cache). */
+#define MVM_TYPE_CACHE_ID_INCR 128
+
 /* S-table, representing a meta-object/representation pairing. Note that the
  * items are grouped in hope that it will pack decently and do decently in
  * terms of cache lines. */
@@ -398,6 +403,16 @@ struct MVMREPROps_Boxing {
     MVMString * (*get_str) (MVMThreadContext *tc, MVMSTable *st,
         MVMObject *root, void *data);
 
+    /* Used with boxing. Sets an unsinged integer value, for representations
+     *  that can hold one. */
+    void (*set_uint) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMuint64 value);
+
+    /* Used with boxing. Gets an unsigned integer value, for representations
+     * that can hold one. */
+    MVMuint64 (*get_uint) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data);
+
     /* Some objects serve primarily as boxes of others, inlining them. This gets
      * gets the reference to such things, using the representation ID to distinguish
      * them. */
@@ -423,10 +438,6 @@ struct MVMREPROps_Positional {
     void (*set_elems) (MVMThreadContext *tc, MVMSTable *st,
         MVMObject *root, void *data, MVMuint64 count);
 
-    /* Returns a true value of the specified index exists, and a false one if not. */
-    MVMint64 (*exists_pos) (MVMThreadContext *tc, MVMSTable *st,
-        MVMObject *root, void *data, MVMint64 index);
-
     /* Pushes the specified value onto the array. */
     void (*push) (MVMThreadContext *tc, MVMSTable *st,
         MVMObject *root, void *data, MVMRegister value, MVMuint16 kind);
@@ -451,6 +462,29 @@ struct MVMREPROps_Positional {
         MVMObject *root, void *data, MVMObject *target_array,
         MVMint64 offset, MVMuint64 elems);
 
+    /* Multi-dimensional array read. */
+    void (*at_pos_multidim) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMint64 num_indices,
+        MVMint64 *indices, MVMRegister *result, MVMuint16 kind);
+
+    /* Multi-dimensional array write. */
+    void (*bind_pos_multidim) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMint64 num_indices,
+        MVMint64 *indices, MVMRegister value, MVMuint16 kind);
+
+    /* Gets the number of dimensions along with a C-level array of them. The
+     * second two parameters are "out"s. The caller must not mutate dimensions,
+     * nor persist it such that it lasts longer than the next VM safepoint. */
+    void (*dimensions) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMint64 *num_dimensions,
+        MVMint64 **dimensions);
+
+    /* Sets the number of dimensions. The caller is responsible for freeing
+     * the array passed in dimensions. */
+    void (*set_dimensions) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *root, void *data, MVMint64 num_dimensions,
+        MVMint64 *dimensions);
+
     /* Gets the STable representing the declared element type. */
     MVMStorageSpec (*get_elem_storage_spec) (MVMThreadContext *tc, MVMSTable *st);
 };
@@ -594,8 +628,8 @@ struct MVMREPROps {
 #define IS_CONCRETE(o)   (!(((MVMObject *)o)->header.flags & MVM_CF_TYPE_OBJECT))
 
 /* Some functions related to 6model core functionality. */
-MVMObject * MVM_6model_get_how(MVMThreadContext *tc, MVMSTable *st);
-MVMObject * MVM_6model_get_how_obj(MVMThreadContext *tc, MVMObject *st);
+MVM_PUBLIC MVMObject * MVM_6model_get_how(MVMThreadContext *tc, MVMSTable *st);
+MVM_PUBLIC MVMObject * MVM_6model_get_how_obj(MVMThreadContext *tc, MVMObject *obj);
 void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name, MVMRegister *res);
 MVM_PUBLIC MVMObject * MVM_6model_find_method_cache_only(MVMThreadContext *tc, MVMObject *obj, MVMString *name);
 MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMString *name,
