@@ -369,6 +369,30 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                             match = 0;
                         break;
                     }
+                    case MVM_SPESH_GUARD_DC_CONC_RW: {
+                        if (STABLE(arg)->container_spec->can_store(tc, arg)) {
+                            MVMRegister dc;
+                            STABLE(arg)->container_spec->fetch(tc, arg, &dc);
+                            if (!dc.o || !IS_CONCRETE(dc.o) || STABLE(dc.o) != st)
+                                match = 0;
+                        }
+                        else {
+                            match = 0;
+                        }
+                        break;
+                    }
+                    case MVM_SPESH_GUARD_DC_TYPE_RW: {
+                        if (STABLE(arg)->container_spec->can_store(tc, arg)) {
+                            MVMRegister dc;
+                            STABLE(arg)->container_spec->fetch(tc, arg, &dc);
+                            if (!dc.o || IS_CONCRETE(dc.o) || STABLE(dc.o) != st)
+                                match = 0;
+                        }
+                        else {
+                            match = 0;
+                        }
+                        break;
+                    }
                     }
                     if (!match)
                         break;
@@ -768,7 +792,7 @@ MVMuint64 MVM_frame_try_return(MVMThreadContext *tc) {
             default:
                 result = NULL;
         }
-        
+
         handler = MVM_frame_find_invokee(tc, hll->exit_handler, NULL);
         two_args_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_TWO_OBJ);
         MVM_args_setup_thunk(tc, NULL, MVM_RETURN_VOID, two_args_callsite);
@@ -1149,12 +1173,18 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
     MVMuint32 ecost = 0;  /* frames traversed with empty cache */
     MVMuint32 xcost = 0;  /* frames traversed with wrong name */
     char *c_name;
+    MVMuint64 start_time;
+    MVMuint64 last_time;
 
     MVMFrame *initial_frame = cur_frame;
     if (!name)
         MVM_exception_throw_adhoc(tc, "Contextual name cannot be null");
-    if (dlog)
+    if (dlog) {
         c_name = MVM_string_utf8_encode_C_string(tc, name);
+        start_time = uv_hrtime();
+        last_time = tc->instance->dynvar_log_lasttime;
+    }
+
     MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names;
@@ -1164,11 +1194,13 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
          * use getdynlex for their own lexicals since the compiler already
          * knows where to find them */
         if (cand && cand->num_inlines) {
-            if (cand->jitcode) {
+            if (cand->jitcode && cur_frame->effective_bytecode == cand->jitcode->bytecode) {
                 void      **labels = cand->jitcode->labels;
                 void *return_label = cur_frame->jit_entry_label;
                 MVMJitInline *inls = cand->jitcode->inlines;
                 MVMint32 i;
+                if (return_label == NULL)
+                    MVM_oops(tc, "Return label is NULL!\n");
                 for (i = 0; i < cand->jitcode->num_inlines; i++) {
                     icost++;
                     if (return_label >= labels[inls[i].start_label] && return_label <= labels[inls[i].end_label]) {
@@ -1185,9 +1217,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                                 if (fcost+icost > 1)
                                   try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                                 if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                                    fprintf(dlog, "I %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                                     fflush(dlog);
                                     MVM_free(c_name);
+                                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                                 }
                                 return result;
                             }
@@ -1213,9 +1246,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                                 if (fcost+icost > 1)
                                   try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                                 if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                                    fprintf(dlog, "I %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                                     fflush(dlog);
                                     MVM_free(c_name);
+                                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                                 }
                                 return result;
                             }
@@ -1235,9 +1269,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                 if (fcost+icost > 5)
                     try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                 if (dlog) {
-                    fprintf(dlog, "C %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                    fprintf(dlog, "C %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                     fflush(dlog);
                     MVM_free(c_name);
+                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                 }
                 return result;
             }
@@ -1257,9 +1292,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                 if (vivify && *type == MVM_reg_obj && !result->o)
                     MVM_frame_vivify_lexical(tc, cur_frame, entry->value);
                 if (dlog) {
-                    fprintf(dlog, "F %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+                    fprintf(dlog, "F %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                     fflush(dlog);
                     MVM_free(c_name);
+                    tc->instance->dynvar_log_lasttime = uv_hrtime();
                 }
                 if (fcost+icost > 1)
                     try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
@@ -1270,9 +1306,10 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
         cur_frame = cur_frame->caller;
     }
     if (dlog) {
-        fprintf(dlog, "N %s %d %d %d %d\n", c_name, fcost, icost, ecost, xcost);
+        fprintf(dlog, "N %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
         fflush(dlog);
         MVM_free(c_name);
+        tc->instance->dynvar_log_lasttime = uv_hrtime();
     }
     return NULL;
 }
@@ -1453,8 +1490,9 @@ static MVMObject * find_invokee_internal(MVMThreadContext *tc, MVMObject *code,
             }
             else {
                 MVMCallsite *new   = MVM_malloc(sizeof(MVMCallsite));
-                MVMint32     fsize = orig->num_pos + (orig->arg_count - orig->num_pos) / 2;
-                new->arg_flags     = MVM_malloc((fsize + 1) * sizeof(MVMCallsiteEntry));
+                MVMint32     fsize = orig->flag_count;
+                new->flag_count    = fsize + 1;
+                new->arg_flags     = MVM_malloc(new->flag_count * sizeof(MVMCallsiteEntry));
                 new->arg_flags[0]  = MVM_CALLSITE_ARG_OBJ;
                 memcpy(new->arg_flags + 1, orig->arg_flags, fsize);
                 new->arg_count      = orig->arg_count + 1;
@@ -1540,6 +1578,9 @@ MVMObject * MVM_frame_context_wrapper(MVMThreadContext *tc, MVMFrame *f) {
             ((MVMContext *)ctx)->body.context = MVM_frame_dec_ref(tc, f);
             ctx = (MVMObject *)MVM_load(&f->context_object);
         }
+        else {
+            f->keep_caller = 1;
+        }
     }
 
     return ctx;
