@@ -37,8 +37,8 @@ typedef struct {
     /* Decode stream, for turning bytes from disk into strings. */
     MVMDecodeStream *ds;
 
-    /* Current separator codepoint. */
-    MVMGrapheme32 sep;
+    /* Current separator specification for line-by-line reading. */
+    MVMDecodeStreamSeparators sep_spec;
 } MVMIOFileData;
 
 /* Closes the file. */
@@ -57,6 +57,18 @@ static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
     return 0;
 }
 
+/* Checks if the file is a TTY. */
+static MVMint64 is_tty(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOFileData *data = (MVMIOFileData *)h->body.data;
+    return (uv_guess_handle(data->fd) == UV_TTY);
+}
+
+/* Gets the file descriptor. */
+static MVMint64 mvm_fileno(MVMThreadContext *tc, MVMOSHandle *h) {
+    MVMIOFileData *data = (MVMIOFileData *)h->body.data;
+    return (MVMint64)data->fd;
+}
+
 /* Sets the encoding used for string-based I/O. */
 static void set_encoding(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 encoding) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
@@ -81,7 +93,7 @@ static void seek(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 offset, MVMint64
         MVM_exception_throw_adhoc(tc, "Failed to seek in filehandle: %d", errno);
     if ((r = MVM_platform_lseek(data->fd, 0, SEEK_CUR)) == -1)
         MVM_exception_throw_adhoc(tc, "Failed to seek in filehandle: %d", errno);
-    data->ds = MVM_string_decodestream_create(tc, data->encoding, r);
+    data->ds = MVM_string_decodestream_create(tc, data->encoding, r, 1);
 }
 
 /* Get curernt position in the file. */
@@ -91,10 +103,9 @@ static MVMint64 mvm_tell(MVMThreadContext *tc, MVMOSHandle *h) {
 }
 
 /* Set the line separator. */
-static void set_separator(MVMThreadContext *tc, MVMOSHandle *h, MVMString *sep) {
+static void set_separator(MVMThreadContext *tc, MVMOSHandle *h, MVMString **seps, MVMint32 num_seps) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
-    data->sep = (MVMGrapheme32)MVM_string_get_grapheme_at(tc, sep,
-        MVM_string_graphs(tc, sep) - 1);
+    MVM_string_decode_stream_sep_from_strings(tc, &(data->sep_spec), seps, num_seps);
 }
 
 /* Read a bunch of bytes into the current decode stream. */
@@ -115,24 +126,26 @@ static MVMint32 read_to_buffer(MVMThreadContext *tc, MVMIOFileData *data, MVMint
 /* Ensures we have a decode stream, creating it if we're missing one. */
 static void ensure_decode_stream(MVMThreadContext *tc, MVMIOFileData *data) {
     if (!data->ds)
-        data->ds = MVM_string_decodestream_create(tc, data->encoding, 0);
+        data->ds = MVM_string_decodestream_create(tc, data->encoding, 0, 1);
 }
 
 /* Reads a single line from the file handle. May serve it from a buffer, if we
  * already read enough data. */
-static MVMString * read_line(MVMThreadContext *tc, MVMOSHandle *h) {
+static MVMString * read_line(MVMThreadContext *tc, MVMOSHandle *h, MVMint32 chomp) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
     ensure_decode_stream(tc, data);
 
     /* Pull data until we can read a line. */
     do {
-        MVMString *line = MVM_string_decodestream_get_until_sep(tc, data->ds, data->sep);
+        MVMString *line = MVM_string_decodestream_get_until_sep(tc,
+            data->ds, &(data->sep_spec), chomp);
         if (line != NULL)
             return line;
     } while (read_to_buffer(tc, data, CHUNK_SIZE) > 0);
 
     /* Reached end of file, or last (non-termianted) line. */
-    return MVM_string_decodestream_get_all(tc, data->ds);
+    return MVM_string_decodestream_get_until_sep_eof(tc, data->ds,
+        &(data->sep_spec), chomp);
 }
 
 /* Reads the file from the current position to the end into a string. */
@@ -146,8 +159,19 @@ static MVMString * slurp(MVMThreadContext *tc, MVMOSHandle *h) {
     if (uv_fs_fstat(tc->loop, &req, data->fd, NULL) < 0) {
         MVM_exception_throw_adhoc(tc, "slurp from filehandle failed: %s", uv_strerror(req.result));
     }
-    while (read_to_buffer(tc, data, req.statbuf.st_size) > 0)
-        ;
+    /* Sometimes - usually for special files like those in /proc - the file
+     * size comes up 0, even though the S_ISREG test succeeds. So in that case
+     * we try a small read and switch to a "read chunks until EOF" impl.
+     * Otherwise we just read the exact size of the file. */
+    if (req.statbuf.st_size == 0) {
+        if (read_to_buffer(tc, data, 32) > 0) {
+            while (read_to_buffer(tc, data, 4096) > 0)
+                ;
+        }
+    } else {
+        while (read_to_buffer(tc, data, req.statbuf.st_size) > 0)
+            ;
+    }
     return MVM_string_decodestream_get_all(tc, data->ds);
 }
 
@@ -204,7 +228,8 @@ static MVMint64 write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMString *str,
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
     MVMuint64 output_size;
     MVMint64 bytes_written;
-    char *output = MVM_string_encode(tc, str, 0, -1, &output_size, data->encoding);
+    char *output = MVM_string_encode(tc, str, 0, -1, &output_size, data->encoding, NULL,
+        MVM_TRANSLATE_NEWLINE_OUTPUT);
     uv_buf_t write_buf  = uv_buf_init(output, output_size);
     uv_fs_t req;
 
@@ -253,6 +278,14 @@ static void truncatefh(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 bytes) {
         MVM_exception_throw_adhoc(tc, "Failed to truncate filehandle: %s", uv_strerror(req.result));
 }
 
+/* Operations aiding process spawning and I/O handling. */
+static void bind_stdio_handle(MVMThreadContext *tc, MVMOSHandle *h, uv_stdio_container_t *stdio,
+        uv_process_t *process) {
+    MVMIOFileData *data = (MVMIOFileData *)h->body.data;
+    stdio->flags        = UV_INHERIT_FD;
+    stdio->data.fd      = data->fd;
+}
+
 /* Locks a file. */
 static MVMint64 lock(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 flag) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
@@ -264,7 +297,7 @@ static MVMint64 lock(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 flag) {
     OVERLAPPED offset;
 
     if (hf == INVALID_HANDLE_VALUE) {
-        MVM_exception_throw_adhoc(tc, "Failed to seek in filehandle: bad file descriptor");
+        MVM_exception_throw_adhoc(tc, "Failed to lock filehandle: bad file descriptor");
     }
 
     flag = ((flag & MVM_FILE_FLOCK_NONBLOCK) ? LOCKFILE_FAIL_IMMEDIATELY : 0)
@@ -304,7 +337,6 @@ static MVMint64 lock(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 flag) {
 
     if (r == -1) {
         MVM_exception_throw_adhoc(tc, "Failed to lock filehandle: %d", errno);
-        return 0;
     }
 
     return 1;
@@ -365,12 +397,15 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
 }
 
 /* IO ops table, populated with functions. */
-static const MVMIOClosable     closable      = { closefh };
-static const MVMIOEncodable    encodable     = { set_encoding };
-static const MVMIOSyncReadable sync_readable = { set_separator, read_line, slurp, read_chars, read_bytes, mvm_eof };
-static const MVMIOSyncWritable sync_writable = { write_str, write_bytes, flush, truncatefh };
-static const MVMIOSeekable     seekable      = { seek, mvm_tell };
-static const MVMIOLockable     lockable      = { lock, unlock };
+static const MVMIOClosable      closable      = { closefh };
+static const MVMIOEncodable     encodable     = { set_encoding };
+static const MVMIOSyncReadable  sync_readable = { set_separator, read_line, slurp, read_chars, read_bytes, mvm_eof };
+static const MVMIOSyncWritable  sync_writable = { write_str, write_bytes, flush, truncatefh };
+static const MVMIOSeekable      seekable      = { seek, mvm_tell };
+static const MVMIOPipeable      pipeable      = { bind_stdio_handle };
+static const MVMIOLockable      lockable      = { lock, unlock };
+static const MVMIOIntrospection introspection = { is_tty, mvm_fileno };
+
 static const MVMIOOps op_table = {
     &closable,
     &encodable,
@@ -378,10 +413,12 @@ static const MVMIOOps op_table = {
     &sync_writable,
     NULL,
     NULL,
+    NULL,
     &seekable,
     NULL,
-    NULL,
+    &pipeable,
     &lockable,
+    &introspection,
     NULL,
     gc_free
 };
@@ -418,7 +455,7 @@ static int resolve_open_mode(int *flag, const char *cp) {
 
 /* Opens a file, returning a synchronous file handle. */
 MVMObject * MVM_file_open_fh(MVMThreadContext *tc, MVMString *filename, MVMString *mode) {
-    char          * const fname  = MVM_string_utf8_encode_C_string(tc, filename);
+    char          * const fname  = MVM_string_utf8_c8_encode_C_string(tc, filename);
     char          * const fmode  = MVM_string_utf8_encode_C_string(tc, mode);
     MVMOSHandle   * const result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
     MVMIOFileData * const data   = MVM_calloc(1, sizeof(MVMIOFileData));
@@ -428,20 +465,23 @@ MVMObject * MVM_file_open_fh(MVMThreadContext *tc, MVMString *filename, MVMStrin
     /* Resolve mode description to flags. */
     int flag;
     if (!resolve_open_mode(&flag, fmode)) {
+        char *waste[] = { fmode, NULL };
         MVM_free(fname);
-        MVM_exception_throw_adhoc(tc, "Invalid open mode: %s", fmode);
+        MVM_exception_throw_adhoc_free(tc, waste, "Invalid open mode: %s", fmode);
     }
     MVM_free(fmode);
 
     /* Try to open the file. */
     if ((fd = uv_fs_open(tc->loop, &req, (const char *)fname, flag, DEFAULT_MODE, NULL)) < 0) {
-        MVM_exception_throw_adhoc(tc, "Failed to open file %s: %s", fname, uv_strerror(req.result));
+        char *waste[] = { fname, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "Failed to open file %s: %s", fname, uv_strerror(req.result));
     }
 
     /* Set up handle. */
     data->fd          = fd;
     data->filename    = fname;
     data->encoding    = MVM_encoding_type_utf8;
+    MVM_string_decode_stream_sep_default(tc, &(data->sep_spec));
     result->body.ops  = &op_table;
     result->body.data = data;
 
