@@ -1,8 +1,6 @@
 #include "moar.h"
 #include "nodes.h"
 
-#include <stdint.h>
-
 /* Some constants. */
 #define HEADER_SIZE                 92
 #define BYTECODE_VERSION            5
@@ -173,7 +171,7 @@ typedef struct {
 } WriterState;
 
 static unsigned int umax(unsigned int a, unsigned int b);
-static void memcpy_endian(char *dest, void *src, size_t size);
+static void memcpy_endian(char *dest, const void *src, size_t size);
 static void write_int64(char *buffer, size_t offset, unsigned long long value);
 static void write_int32(char *buffer, size_t offset, unsigned int value);
 static void write_int16(char *buffer, size_t offset, unsigned short value);
@@ -198,7 +196,7 @@ static unsigned int umax(unsigned int a, unsigned int b) {
 }
 
 /* copies memory dependent on endianness */
-static void memcpy_endian(char *dest, void *src, size_t size) {
+static void memcpy_endian(char *dest, const void *src, size_t size) {
 #ifdef MVM_BIGENDIAN
     size_t i;
     char *srcbytes = (char *)src;
@@ -323,18 +321,35 @@ static unsigned short type_to_local_type(VM, WriterState *ws, MASTNode *type) {
     if (ss->inlineable) {
         switch (ss->boxed_primitive) {
             case MVM_STORAGE_SPEC_BP_INT:
-                switch (ss->bits) {
-                    case 8:
-                        return MVM_reg_int8;
-                    case 16:
-                        return MVM_reg_int16;
-                    case 32:
-                        return MVM_reg_int32;
-                    case 64:
-                        return MVM_reg_int64;
-                    default:
-                        cleanup_all(vm, ws);
-                        DIE(vm, "Invalid int size for local/lexical");
+                if (ss->is_unsigned) {
+                    switch (ss->bits) {
+                        case 8:
+                            return MVM_reg_uint8;
+                        case 16:
+                            return MVM_reg_uint16;
+                        case 32:
+                            return MVM_reg_uint32;
+                        case 64:
+                            return MVM_reg_uint64;
+                        default:
+                            cleanup_all(vm, ws);
+                            DIE(vm, "Invalid int size for local/lexical");
+                    }
+                }
+                else {
+                    switch (ss->bits) {
+                        case 8:
+                            return MVM_reg_int8;
+                        case 16:
+                            return MVM_reg_int16;
+                        case 32:
+                            return MVM_reg_int32;
+                        case 64:
+                            return MVM_reg_int64;
+                        default:
+                            cleanup_all(vm, ws);
+                            DIE(vm, "Invalid int size for local/lexical");
+                    }
                 }
                 break;
             case MVM_STORAGE_SPEC_BP_NUM:
@@ -350,7 +365,6 @@ static unsigned short type_to_local_type(VM, WriterState *ws, MASTNode *type) {
                 break;
             case MVM_STORAGE_SPEC_BP_STR:
                 return MVM_reg_str;
-                break;
             default:
                 cleanup_all(vm, ws);
                 DIE(vm, "Type used for local/lexical has invalid boxed primitive in storage spec");
@@ -1140,12 +1154,12 @@ static void compile_frame(VM, WriterState *ws, MASTNode *node, unsigned short id
                 ((MAST_Frame *)f->outer)->index);
         }
         else {
-            unsigned short i, found, num_frames;
+            unsigned short j, found, num_frames;
             found = 0;
             num_frames = (unsigned short)ELEMS(vm, ws->cu->frames);
-            for (i = 0; i < num_frames; i++) {
-                if (ATPOS(vm, ws->cu->frames, i) == f->outer) {
-                    write_int16(ws->frame_seg, ws->frame_pos + 24, i);
+            for (j = 0; j < num_frames; j++) {
+                if (ATPOS(vm, ws->cu->frames, j) == f->outer) {
+                    write_int16(ws->frame_seg, ws->frame_pos + 24, j);
                     found = 1;
                     break;
                 }
@@ -1271,9 +1285,9 @@ static void compile_frame(VM, WriterState *ws, MASTNode *node, unsigned short id
         write_int16(ws->frame_seg, ws->frame_pos + 2,
             (MVMuint16)ATPOS_I(vm, f->static_lex_values, 4 * i + 1));
         write_int32(ws->frame_seg, ws->frame_pos + 4,
-            (MVMuint16)ATPOS_I(vm, f->static_lex_values, 4 * i + 2));
+            (MVMuint32)ATPOS_I(vm, f->static_lex_values, 4 * i + 2));
         write_int32(ws->frame_seg, ws->frame_pos + 8,
-            (MVMuint16)ATPOS_I(vm, f->static_lex_values, 4 * i + 3));
+            (MVMuint32)ATPOS_I(vm, f->static_lex_values, 4 * i + 3));
         ws->frame_pos += FRAME_SLV_SIZE;
     }
 
@@ -1317,13 +1331,15 @@ static char * form_string_heap(VM, WriterState *ws, unsigned int *string_heap_si
         unsigned short align;
         unsigned int need;
 
-        /* Decide if we can get away with Latin-1. */
+        /* Decide if we can get away with Latin-1 with an assumption of the
+         * string already being in NFG. Latin-1 is except \r, which we also
+         * check for here. */
         MVMint32   need_utf8 = 0;
         MVMString *str       = ATPOS_S(vm, ws->strings, i);
         MVM_string_gi_init(tc, &gi, str);
         while (MVM_string_gi_has_more(tc, &gi)) {
             MVMGrapheme32 g = MVM_string_gi_get_grapheme(tc, &gi);
-            if (g < 0 || g >= 0xFF) {
+            if (g < 0 || g >= 0xFF || g == 0x0D) {
                 need_utf8 = 1;
                 break;
             }
@@ -1331,8 +1347,8 @@ static char * form_string_heap(VM, WriterState *ws, unsigned int *string_heap_si
 
         /* Encode it with the chosen algorithm. */
         encoded = need_utf8
-            ? MVM_string_utf8_encode(tc, str, &bytelen)
-            : MVM_string_latin1_encode(tc, str, &bytelen);
+            ? MVM_string_utf8_encode(tc, str, &bytelen, 0)
+            : MVM_string_latin1_encode(tc, str, &bytelen, 0);
         if (bytelen > 0x3FFFFFFF) {
             cleanup_all(vm, ws);
             DIE(vm, "String too long for string constants segment");
@@ -1370,9 +1386,9 @@ static char * form_string_heap(VM, WriterState *ws, unsigned int *string_heap_si
 
 /* Takes all the pieces and forms the bytecode output. */
 static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_size) {
-    unsigned int size    = 0;
-    unsigned int pos     = 0;
-    char         *output = NULL;
+    MVMuint32     size    = 0;
+    MVMuint32     pos     = 0;
+    char         *output;
     unsigned int  string_heap_size;
     char         *string_heap;
     unsigned int  hll_str_idx;
@@ -1387,16 +1403,16 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
     string_heap = form_string_heap(vm, ws, &string_heap_size);
 
     /* Work out total size. */
-    size += HEADER_SIZE;
-    size += string_heap_size;
-    size += ws->scdep_bytes;
-    size += ws->extops_bytes;
-    size += ws->frame_pos;
-    size += ws->callsite_pos;
-    size += ws->bytecode_pos;
-    size += ws->annotation_pos;
+    size += MVM_ALIGN_SECTION(HEADER_SIZE);
+    size += MVM_ALIGN_SECTION(string_heap_size);
+    size += MVM_ALIGN_SECTION(ws->scdep_bytes);
+    size += MVM_ALIGN_SECTION(ws->extops_bytes);
+    size += MVM_ALIGN_SECTION(ws->frame_pos);
+    size += MVM_ALIGN_SECTION(ws->callsite_pos);
+    size += MVM_ALIGN_SECTION(ws->bytecode_pos);
+    size += MVM_ALIGN_SECTION(ws->annotation_pos);
     if (vm->serialized)
-        size += vm->serialized_size;
+        size += MVM_ALIGN_SECTION(vm->serialized_size);
 
     /* Allocate space for the bytecode output. */
     output = (char *)MVM_malloc(size);
@@ -1405,37 +1421,37 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
     /* Generate start of header. */
     memcpy(output, "MOARVM\r\n", 8);
     write_int32(output, 8, BYTECODE_VERSION);
-    pos += HEADER_SIZE;
+    pos += MVM_ALIGN_SECTION(HEADER_SIZE);
 
     /* Add SC dependencies section and its header entries. */
     write_int32(output, SCDEP_HEADER_OFFSET, pos);
     write_int32(output, SCDEP_HEADER_OFFSET + 4, ELEMS(vm, ws->cu->sc_handles));
     memcpy(output + pos, ws->scdep_seg, ws->scdep_bytes);
-    pos += ws->scdep_bytes;
+    pos += MVM_ALIGN_SECTION(ws->scdep_bytes);
 
     /* Add extension ops section and its header entries. */
     write_int32(output, EXTOP_HEADER_OFFSET, pos);
     write_int32(output, EXTOP_HEADER_OFFSET + 4, ws->num_extops);
     memcpy(output + pos, ws->extops_seg, ws->extops_bytes);
-    pos += ws->extops_bytes;
+    pos += MVM_ALIGN_SECTION(ws->extops_bytes);
 
     /* Add frames section and its header entries. */
     write_int32(output, FRAME_HEADER_OFFSET, pos);
     write_int32(output, FRAME_HEADER_OFFSET + 4, ws->num_frames);
     memcpy(output + pos, ws->frame_seg, ws->frame_pos);
-    pos += ws->frame_pos;
+    pos += MVM_ALIGN_SECTION(ws->frame_pos);
 
     /* Add callsites section and its header entries. */
     write_int32(output, CALLSITE_HEADER_OFFSET, pos);
     write_int32(output, CALLSITE_HEADER_OFFSET + 4, ws->num_callsites);
     memcpy(output + pos, ws->callsite_seg, ws->callsite_pos);
-    pos += ws->callsite_pos;
+    pos += MVM_ALIGN_SECTION(ws->callsite_pos);
 
     /* Add strings heap section and its header entries. */
     write_int32(output, STRING_HEADER_OFFSET, pos);
     write_int32(output, STRING_HEADER_OFFSET + 4, ELEMS(vm, ws->strings));
     memcpy(output + pos, string_heap, string_heap_size);
-    pos += string_heap_size;
+    pos += MVM_ALIGN_SECTION(string_heap_size);
     if (string_heap) {
         MVM_free(string_heap);
         string_heap = NULL;
@@ -1446,7 +1462,7 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
         write_int32(output, SCDATA_HEADER_OFFSET, pos);
         write_int32(output, SCDATA_HEADER_OFFSET + 4, vm->serialized_size);
         memcpy(output + pos, vm->serialized, vm->serialized_size);
-        pos += vm->serialized_size;
+        pos += MVM_ALIGN_SECTION(vm->serialized_size);
         MVM_free(vm->serialized);
         vm->serialized = NULL;
         vm->serialized_size = 0;
@@ -1456,13 +1472,13 @@ static char * form_bytecode_output(VM, WriterState *ws, unsigned int *bytecode_s
     write_int32(output, BYTECODE_HEADER_OFFSET, pos);
     write_int32(output, BYTECODE_HEADER_OFFSET + 4, ws->bytecode_pos);
     memcpy(output + pos, ws->bytecode_seg, ws->bytecode_pos);
-    pos += ws->bytecode_pos;
+    pos += MVM_ALIGN_SECTION(ws->bytecode_pos);
 
     /* Add annotation section and its header entries (offset, length). */
     write_int32(output, ANNOTATION_HEADER_OFFSET, pos);
     write_int32(output, ANNOTATION_HEADER_OFFSET + 4, ws->annotation_pos);
     memcpy(output + pos, ws->annotation_seg, ws->annotation_pos);
-    pos += ws->annotation_pos;
+    pos += MVM_ALIGN_SECTION(ws->annotation_pos);
 
     /* Add HLL and special frame indexes. */
     write_int32(output, HLL_NAME_HEADER_OFFSET, hll_str_idx);
@@ -1536,7 +1552,7 @@ char * MVM_mast_compile(VM, MASTNode *node, MASTNodeTypes *types, unsigned int *
     if (vm->serialized_string_heap) {
         MVMint64 elems = ELEMS(vm, vm->serialized_string_heap);
         for (i = 1; i < elems; i++)
-            get_string_heap_index(vm, ws, ATPOS_S(vm, vm->serialized_string_heap, i));
+            (void)get_string_heap_index(vm, ws, ATPOS_S(vm, vm->serialized_string_heap, i));
         vm->serialized_string_heap = NULL;
     }
 
