@@ -51,6 +51,24 @@ static void create_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 obj_o
     }
 }
 
+static void create_facts_with_type(MVMThreadContext *tc, MVMSpeshGraph *g,
+                                   MVMuint16 obj_orig, MVMuint16 obj_i,
+                                   MVMObject *type) {
+    MVMSpeshFacts *obj_facts  = &(g->facts[obj_orig][obj_i]);
+
+    /* The type is carried. */
+    obj_facts->type   = type;
+    obj_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+
+    /* We know it's a concrete object. */
+    obj_facts->flags |= MVM_SPESH_FACT_CONCRETE;
+
+    /* If we know the type object, then we can check to see if
+     * it's a container type. */
+    if (type && !STABLE(type)->container_spec)
+        obj_facts->flags |= MVM_SPESH_FACT_DECONTED;
+}
+
 /* Adds facts from knowing the exact value being put into an object local. */
 static void object_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_orig,
                          MVMuint16 tgt_i, MVMObject *obj) {
@@ -257,23 +275,59 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MV
     /* Produce a guard op and set facts. */
     if (stable_cont) {
         MVMSpeshOperand reg  = ins->operands[0];
+        MVMContainerSpec *cs = (MVMContainerSpec *) STABLE(stable_cont)->container_spec;
         facts                = &g->facts[reg.reg.orig][reg.reg.i];
         facts->type          = STABLE(stable_cont)->WHAT;
         facts->flags        |= (MVM_SPESH_FACT_KNOWN_TYPE | MVM_SPESH_FACT_CONCRETE |
                                MVM_SPESH_FACT_KNOWN_DECONT_TYPE);
         facts->decont_type   = STABLE(stable_value)->WHAT;
-        if (IS_CONCRETE(stable_value)) {
-            facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
-            ins->info = MVM_op_get_op(MVM_OP_sp_guardcontconc);
+
+        /* If this is a native container, we get away with testing
+         * against the STABLE only, as the NativeRef REPR has all
+         * interesting values in its REPRData. */
+        if (cs->can_store(tc, stable_cont) &&
+                (MVM_6model_container_iscont_i(tc, stable_cont) ||
+                MVM_6model_container_iscont_n(tc, stable_cont) ||
+                MVM_6model_container_iscont_s(tc, stable_cont))) {
+            facts         = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
+            /*facts->type   = STABLE(stable_value)->WHAT;*/
+            facts->flags |= MVM_SPESH_FACT_RW_CONT;
+
+            ins->info = MVM_op_get_op(MVM_OP_sp_guardconc);
+
+            ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
+            ins->operands[0] = reg;
+            ins->operands[1].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_cont));
+        } else {
+            if (cs->can_store(tc, stable_cont)) {
+                /* We could do stability testing on rw-ness too, but it's quite
+                 * unlikely we'll have codepaths with a mix of readable and
+                 * writable containers. */
+                facts->flags |= MVM_SPESH_FACT_RW_CONT;
+                if (IS_CONCRETE(stable_value)) {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardrwconc);
+                }
+                else {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardrwtype);
+                }
+            }
+            else {
+                if (IS_CONCRETE(stable_value)) {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_CONCRETE;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardcontconc);
+                }
+                else {
+                    facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
+                    ins->info = MVM_op_get_op(MVM_OP_sp_guardconttype);
+                }
+            }
+            ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+            ins->operands[0] = reg;
+            ins->operands[1].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_cont));
+            ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_value));
         }
-        else {
-            facts->flags |= MVM_SPESH_FACT_DECONT_TYPEOBJ;
-            ins->info = MVM_op_get_op(MVM_OP_sp_guardconttype);
-        }
-        ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
-        ins->operands[0] = reg;
-        ins->operands[1].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_cont));
-        ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)STABLE(stable_value));
     }
     else {
         facts         = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
@@ -467,6 +521,11 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[1].reg.orig, ins->operands[1].reg.i);
             break;
+        case MVM_OP_newexception:
+            create_facts_with_type(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                tc->instance->boot_types.BOOTException);
+            break;
         case MVM_OP_const_i64:
         case MVM_OP_const_i32:
         case MVM_OP_const_i16:
