@@ -6,13 +6,13 @@
  * in http://www.cs.rice.edu/~keith/EMBED/dom.pdf. The SSA algorithm itself is
  * from http://www.cs.utexas.edu/~pingali/CS380C/2010/papers/ssaCytron.pdf. */
 
-#define GET_I8(pc, idx)     *((MVMint8 *)(pc + idx))
-#define GET_UI8(pc, idx)    *((MVMuint8 *)(pc + idx))
-#define GET_I16(pc, idx)    *((MVMint16 *)(pc + idx))
-#define GET_UI16(pc, idx)   *((MVMuint16 *)(pc + idx))
-#define GET_I32(pc, idx)    *((MVMint32 *)(pc + idx))
-#define GET_UI32(pc, idx)   *((MVMuint32 *)(pc + idx))
-#define GET_N32(pc, idx)    *((MVMnum32 *)(pc + idx))
+#define GET_I8(pc, idx)     *((MVMint8 *)((pc) + (idx)))
+#define GET_UI8(pc, idx)    *((MVMuint8 *)((pc) + (idx)))
+#define GET_I16(pc, idx)    *((MVMint16 *)((pc) + (idx)))
+#define GET_UI16(pc, idx)   *((MVMuint16 *)((pc) + (idx)))
+#define GET_I32(pc, idx)    *((MVMint32 *)((pc) + (idx)))
+#define GET_UI32(pc, idx)   *((MVMuint32 *)((pc) + (idx)))
+#define GET_N32(pc, idx)    *((MVMnum32 *)((pc) + (idx)))
 
 /* Allocate a piece of memory from the spesh graph's buffer. Deallocated when
  * the spesh graph is. */
@@ -258,6 +258,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                         (int)type, ins_node->info->name);
                 }
                 break;
+            default:
+                break;
             }
             }
         }
@@ -298,6 +300,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
             next_bbs = 1;
             break;
+        default:
+            break;
         }
 
         /* Final instruction is basic block end. */
@@ -588,7 +592,8 @@ static void add_predecessors(MVMThreadContext *tc, MVMSpeshGraph *g) {
             MVMSpeshBB  *tgt = cur_bb->succ[i];
             MVMSpeshBB **new_pred = MVM_spesh_alloc(tc, g,
                 (tgt->num_pred + 1) * sizeof(MVMSpeshBB *));
-            memcpy(new_pred, tgt->pred, tgt->num_pred * sizeof(MVMSpeshBB *));
+            if (tgt->num_pred)
+                memcpy(new_pred, tgt->pred, tgt->num_pred * sizeof(MVMSpeshBB *));
             new_pred[tgt->num_pred] = cur_bb;
             tgt->pred = new_pred;
             tgt->num_pred++;
@@ -727,7 +732,8 @@ static void add_child(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *target
 
     /* Nope, so insert. */
     new_children = MVM_spesh_alloc(tc, g, (target->num_children + 1) * sizeof(MVMSpeshBB *));
-    memcpy(new_children, target->children, target->num_children * sizeof(MVMSpeshBB *));
+    if (target->num_children)
+        memcpy(new_children, target->children, target->num_children * sizeof(MVMSpeshBB *));
     new_children[target->num_children] = to_add;
     target->children = new_children;
     target->num_children++;
@@ -754,7 +760,8 @@ static void add_to_frontier_set(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpesh
 
     /* Nope, so insert. */
     new_df = MVM_spesh_alloc(tc, g, (target->num_df + 1) * sizeof(MVMSpeshBB *));
-    memcpy(new_df, target->df, target->num_df * sizeof(MVMSpeshBB *));
+    if (target->num_df)
+        memcpy(new_df, target->df, target->num_df * sizeof(MVMSpeshBB *));
     new_df[target->num_df] = to_add;
     target->df = new_df;
     target->num_df++;
