@@ -1,7 +1,8 @@
 /* Normalization modes. Numbers picked so that:
  *  - The LSB tells us whether to do canonical or compatibility normalization
  *  - The second bit tells us whether to do canonical normalization
- *  - The third bit tells us to go a step further and create synthetic codes.
+ *  - The third bit tells us to go a step further and create synthetic codes
+ *    for graphemes.
  */ 
 typedef enum {
     MVM_NORMALIZE_NFD   = 0,
@@ -23,6 +24,9 @@ typedef enum {
 #define MVM_NORMALIZE_FIRST_SIG_NFKD    0x00A0
 #define MVM_NORMALIZE_FIRST_SIG_NFKC    0x00A0
 
+/* First codepoint with a non-zero canonical combining class. */
+#define MVM_NORMALIZE_FIRST_NONZERO_CCC 0x300
+
 /* Streaming Unicode normalizer structure. */
 struct MVMNormalizer {
     /* What form of normalization are we doing? */
@@ -50,6 +54,10 @@ struct MVMNormalizer {
 
     /* The quickcheck property for the normalization form in question. */
     MVMint32 quick_check_property;
+
+    /* If we should translate the \r\n grapheme to \n (only applicable when
+     * normalizing to NFG). */
+    MVMint32 translate_newlines;
 };
 
 /* Guts-y functions, called by the API level ones below. */
@@ -62,10 +70,14 @@ MVMint32 MVM_unicode_normalizer_process_codepoint_norm_terminator(MVMThreadConte
  * codepoints now available including the one we just passed out. If we can't
  * produce a normalized codepoint right now, we return a 0. */
 MVM_STATIC_INLINE MVMint32 MVM_unicode_normalizer_process_codepoint(MVMThreadContext *tc, MVMNormalizer *n, MVMCodepoint in, MVMCodepoint *out) {
-    /* If we have \n then we always treat this as a "sequence point" in the
-     * normalization process. */
-    if (in == '\n')
-        return MVM_unicode_normalizer_process_codepoint_norm_terminator(tc, n, in, out);
+    /* Control characters in the Latin-1 range are normalization terminators -
+     * that is, we know we can spit out whatever codepoints we have seen so
+     * far in normalized form without having to consider them into the
+     * normalization process. The exception is if we're computing NFG, and
+     * we got \r, which can form a grapheme in the case of \r\n. */
+    if (in < 0x20 || in >= 0x7F && in <= 0x9F || in == 0xAD)
+        if (!(MVM_NORMALIZE_GRAPHEME(n->form) && in == 0x0D))
+            return MVM_unicode_normalizer_process_codepoint_norm_terminator(tc, n, in, out);
 
     /* Fast-paths apply when the codepoint to consider is too low to have any
      * interesting properties in the target normalization form. */
@@ -73,12 +85,15 @@ MVM_STATIC_INLINE MVMint32 MVM_unicode_normalizer_process_codepoint(MVMThreadCon
         if (MVM_NORMALIZE_COMPOSE(n->form)) {
             /* For the composition fast path we always have to know that we've
             * seen two codepoints in a row that are below those needing a full
-            * check. Then we can spit out the first one. */
-            if (n->buffer_end - n->buffer_start == 1) {
-                if (n->buffer[n->buffer_start] < n->first_significant) {
-                    *out = n->buffer[n->buffer_start];
-                    n->buffer[n->buffer_start] = in;
-                    return 1;
+            * check. Then we can spit out the first one. Exception: we are
+            * normalizing to graphemes and see \r. */
+            if (!(MVM_NORMALIZE_GRAPHEME(n->form) && in == 0x0D)) {
+                if (n->buffer_end - n->buffer_start == 1) {
+                    if (n->buffer[n->buffer_start] < n->first_significant) {
+                        *out = n->buffer[n->buffer_start];
+                        n->buffer[n->buffer_start] = in;
+                        return 1;
+                    }
                 }
             }
         }
@@ -104,6 +119,9 @@ MVM_STATIC_INLINE MVMint32 MVM_unicode_normalizer_process_codepoint_to_grapheme(
     return MVM_unicode_normalizer_process_codepoint(tc, n, in, (MVMGrapheme32 *)out);
 }
 
+/* Push a number of codepoints into the "to normalize" buffer. */
+void MVM_unicode_normalizer_push_codepoints(MVMThreadContext *tc, MVMNormalizer *n, const MVMCodepoint *in, MVMint32 num_codepoints);
+
 /* Get the number of codepoints/graphemes ready to fetch. */
 MVM_STATIC_INLINE MVMint32 MVM_unicode_normalizer_available(MVMThreadContext *tc, MVMNormalizer *n) {
     return n->buffer_norm_end - n->buffer_start;
@@ -135,15 +153,16 @@ MVM_STATIC_INLINE MVMGrapheme32 MVM_unicode_normalizer_get_grapheme(MVMThreadCon
 /* Setup and teardown of the MVMNormalizer struct. */
 MVMNormalization MVN_unicode_normalizer_form(MVMThreadContext *tc, MVMint64 form_in);
 void MVM_unicode_normalizer_init(MVMThreadContext *tc, MVMNormalizer *n, MVMNormalization norm);
+void MVM_unicode_normalizer_translate_newlines(MVMThreadContext *tc, MVMNormalizer *n);
 void MVM_unicode_normalizer_cleanup(MVMThreadContext *tc, MVMNormalizer *n);
 
 /* High-level normalize implementation, working from an input array of
  * codepoints and producing an output array of codepoints. */
-void MVM_unicode_normalize_codepoints(MVMThreadContext *tc, MVMObject *in, MVMObject *out, MVMNormalization form);
+void MVM_unicode_normalize_codepoints(MVMThreadContext *tc, const MVMObject *in, MVMObject *out, MVMNormalization form);
 
 /* High-level function to produces an NFG string from an input array of
  * codepoints. */
-MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, MVMObject *codes);
+MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc,const MVMObject *codes);
 
 /* High-level function to produce an array of codepoints from a string. */
 void MVM_unicode_string_to_codepoints(MVMThreadContext *tc, MVMString *s, MVMNormalization form, MVMObject *out);
